@@ -1,26 +1,28 @@
 """Callbacks that can be added to a model trainer's main loop."""
-# TODO: implement experiment logger
-# TODO: implement csv logger
-
+import os
 import abc
 import logging
 import random
 import string
 import subprocess
+import datetime
 import time
 import numpy as np
+import pandas as pd
 
 from tqdm import tqdm
 from torchvision.utils import make_grid
 import visdom
 
 from .utils import ExponentialMovingAverage
+from .database import SQLiteDatabase
 
 
 __all__ = [
     "Callback",
     "CheckpointingCallback",
     "LoggingCallback",
+    "SQLLoggingCallback",
     "ImageDisplayCallback",
     "ProgressBarCallback",
     "VisdomLoggingCallback",
@@ -32,7 +34,15 @@ LOG = logging.getLogger(__name__)
 
 
 class Callback(object):
-    """Base class for all training callbacks."""
+    """Base class for all training callbacks.
+
+    Attributes:
+        epoch(int): current epoch index.
+        batch(int): current batch index.
+        datasize(int): number of batches in the training dataset.
+        val_datasize(int): number of batches in the validation dataset.
+        model_interface(ttools.ModelInterface): parent interface driving the training.
+    """
 
     def __repr__(self):
         return self.__class__.__name__
@@ -41,18 +51,30 @@ class Callback(object):
         super(Callback, self).__init__()
         self.epoch = 0
         self.batch = 0
+        self.val_batch = 0
         self.datasize = 0
         self.val_datasize = 0
+        self.model_interface = None
 
     def training_start(self, dataloader):
+        """Hook to execute code when the training begins.
+
+        Args:
+            dataloader(th.utils.data.Dataloader): a data loading class that
+            provides batches of data for training.
+        """
         self.datasize = len(dataloader)
 
     def training_end(self):
+        """Hook to execute code when the training ends."""
         pass
 
     def epoch_start(self, epoch):
         """Hook to execute code when a new epoch starts.
 
+        Args:
+            epoch(int): index of the current epoch.
+
         Note: self.epoch is never incremented. Instead, it should be set by the
         caller.
         """
@@ -61,24 +83,60 @@ class Callback(object):
     def epoch_end(self):
         """Hook to execute code when an epoch ends.
 
-        Note: self.epoch is never incremented, but it is set externally in
-        `epoch_start`.
+        NOTE: self.epoch is not incremented. Instead it is set externally in
+        the `epoch_start` method.
         """
         pass
 
     def validation_start(self, dataloader):
+        """Hook to execute code when a validation run starts.
+
+        Args:
+            dataloader(th.utils.data.Dataloader): a data loading class that
+            provides batches of data for evaluation.
+        """
         self.val_datasize = len(dataloader)
 
-    def validation_step(self, batch, fwd_data, val_data):
-        pass
-
     def validation_end(self, val_data):
+        """Hook to execute code when a validation run ends."""
         pass
 
-    def batch_start(self, batch, batch_data):
-        self.batch = batch
+    def batch_start(self, batch_idx, batch_data):
+        """Hook to execute code when a training step starts.
 
-    def batch_end(self, batch_data, fwd_result, bwd_result):
+        Args:
+            batch_idx(int): index of the current batch.
+            batch_data: a Tensor, tuple of dict with the current batch of data.
+        """
+        self.batch = batch_idx
+
+    def batch_end(self, batch_data, train_step_data):
+        """Hook to execute code when a training step ends.
+
+        Args:
+            batch_data: a Tensor, tuple of dict with the current batch of data.
+            train_setp_data(dict): outputs from the `training_step` of a
+                ModelInterface.
+        """
+        pass
+
+    def val_batch_start(self, batch_idx, batch_data):
+        """Hook to execute code when a validation step starts.
+
+        Args:
+            batch_idx(int): index of the current batch.
+            batch_data: a Tensor, tuple of dict with the current batch of data.
+        """
+        self.val_batch = batch_idx
+
+    def val_batch_end(self, batch_data, running_val_data):
+        """Hook to execute code when a validation step ends.
+
+        Args:
+            batch_data: a Tensor, tuple of dict with the current batch of data.
+            train_setp_data(dict): running outputs produced by the `validation_step` of a
+                ModelInterface.
+        """
         pass
 
 
@@ -89,29 +147,32 @@ class KeyedCallback(Callback):
     validation_data) produced by a ModelInterface.
 
     Args:
-      keys (list of str): list of keys whose values will be logged during training. Defaults
-        to ["loss"].
-      val_keys (list of str): list of keys whose values will be logged during validation
+      keys (list of str or None): list of keys whose values will be logged during
+          training.
+      val_keys (list of str or None): list of keys whose values will be logged during
+          validation
     """
-
-    def __init__(self, keys=None, val_keys=None, smoothing=0.99):
+    def __init__(self, keys=None, val_keys=None, smoothing=0.999):
         super(KeyedCallback, self).__init__()
+        if keys is None and val_keys is None:
+            LOG.warning("Logger has no keys, nor val_keys")
+
         if keys is None:
-            self.keys = ["loss"]
+            self.keys = []
         else:
             self.keys = keys
 
         if val_keys is None:
-            self.val_keys = [] 
-            # self.val_keys = self.keys
+            self.val_keys = []
         else:
             self.val_keys = val_keys
 
+        # Only smooth the training keys
         self.ema = ExponentialMovingAverage(self.keys, alpha=smoothing)
 
-    def batch_end(self, batch_data, fwd, bwd):
+    def batch_end(self, batch_data, train_step_data):
         for k in self.keys:
-            self.ema.update(k, bwd[k])
+            self.ema.update(k, train_step_data[k])
 
 
 class VisdomLoggingCallback(KeyedCallback):
@@ -128,7 +189,7 @@ class VisdomLoggingCallback(KeyedCallback):
         0.0 disables smoothing.
     """
 
-    def __init__(self, keys=None, val_keys=None, frequency=100, server=None, 
+    def __init__(self, keys=None, val_keys=None, frequency=100, server=None,
                  port=8097, base_url="/", env="main", log=False, smoothing=0.99):
         super(VisdomLoggingCallback, self).__init__(
             keys=keys, val_keys=val_keys, smoothing=smoothing)
@@ -157,19 +218,19 @@ class VisdomLoggingCallback(KeyedCallback):
         self._step = 0
         self.frequency = frequency
 
-    def batch_end(self, batch_data, fwd, bwd):
-        super(VisdomLoggingCallback, self).batch_end(batch_data, fwd, bwd)
+    def batch_end(self, batch_data, train_step_data):
+        super(VisdomLoggingCallback, self).batch_end(batch_data, train_step_data)
 
         if self._step % self.frequency != 0:
             self._step += 1
             return
         self._step = 0
 
-        t = self.batch / self.datasize + self.epoch
+        t = self.batch / max(self.datasize, 1) + self.epoch
 
         for k in self.keys:
-            self._api.line([self.ema[k]], [t], update="append", win=k, name="train",
-                           opts=self._opts[k])
+            self._api.line([self.ema[k]], [t], update="append", win=k,
+                           name="train", opts=self._opts[k])
 
         self._step += 1
 
@@ -214,35 +275,29 @@ class MultiPlotCallback(KeyedCallback):
             self._api.close(win)
 
         # Cleanup previous plots and setup options
-        all_keys = set(self.keys + self.val_keys)
-        legend = []
-        for k in list(all_keys):
-            if k in self.keys:
-                legend.append(k)
-            if k in self.val_keys:
-                legend.append(k + "_val")
+        legend = self.keys
 
         self._opts = {
             "legend": legend,
             "title": self.win,
             "xlabel": "epoch",
         }
+
         if log:
             self._opts["ytype"] = "log"
 
         self._step = 0
         self.frequency = frequency
 
-    def batch_end(self, batch_data, fwd, bwd):
-        super(MultiPlotCallback, self).batch_end(batch_data, fwd, bwd)
+    def batch_end(self, batch_data, train_step_data):
+        super(MultiPlotCallback, self).batch_end(batch_data, train_step_data)
 
         if self._step % self.frequency != 0:
             self._step += 1
             return
         self._step = 0
 
-        t = self.batch / self.datasize + self.epoch
-
+        t = self.batch / max(self.datasize, 1) + self.epoch
 
         data = np.array([self.ema[k] for k in self.keys])
         data = np.expand_dims(data, 1)
@@ -250,8 +305,8 @@ class MultiPlotCallback(KeyedCallback):
 
         self._step += 1
 
-    def validation_end(self, val_data):
-        pass
+    # def validation_end(self, val_data):
+    #     pass
         # super(MultiPlotCallback, self).validation_end(val_data)
         # t = self.epoch + 1
         # for k in self.val_keys:
@@ -319,9 +374,9 @@ class LoggingCallback(KeyedCallback):
         self.__print(s)
         self.__unindent()
 
-    def batch_end(self, batch_data, fwd, bwd_data):
+    def batch_end(self, batch_data, train_step_data):
         """Logs training advancement Epoch.Batch"""
-        super(LoggingCallback, self).batch_end(batch_data, fwd, bwd_data)
+        super(LoggingCallback, self).batch_end(batch_data, train_step_data)
 
         if self._step % self.frequency != 0:
             self._step += 1
@@ -330,8 +385,7 @@ class LoggingCallback(KeyedCallback):
 
         s = "Step {}.{}".format(self.epoch + 1, self.batch + 1)
         for k in self.keys:
-            print(bwd_data)
-            value = bwd_data[k]
+            value = train_step_data[k]
             if value is not None:
                 s += " | {} = {:.2f}".format(k, value)
         self.__print(s)
@@ -342,9 +396,14 @@ class LoggingCallback(KeyedCallback):
 class ProgressBarCallback(KeyedCallback):
     """A progress bar optimization logger."""
 
-    def __init__(self, keys=None, val_keys=None, smoothing=0.99):
-        super(ProgressBarCallback, self).__init__(keys=keys, val_keys=val_keys, smoothing=smoothing)
+    def __init__(self, keys=None, val_keys=None, smoothing=0.99, label=None):
+        super(ProgressBarCallback, self).__init__(
+            keys=keys, val_keys=val_keys, smoothing=smoothing)
         self.pbar = None
+        if label is None:
+            self.label = ""
+        else:
+            self.label = label
 
     def training_start(self, dataloader):
         super(ProgressBarCallback, self).training_start(dataloader)
@@ -356,8 +415,11 @@ class ProgressBarCallback(KeyedCallback):
 
     def epoch_start(self, epoch):
         super(ProgressBarCallback, self).epoch_start(epoch)
+        desc = "Epoch {}".format(self.epoch)
+        if self.label is not None:
+            desc = "%s | " % self.label + desc
         self.pbar = tqdm(total=self.datasize, unit=" batches",
-                         desc="Epoch {}".format(self.epoch + 1))
+                         desc=desc)
 
     def epoch_end(self):
         super(ProgressBarCallback, self).epoch_end()
@@ -368,9 +430,9 @@ class ProgressBarCallback(KeyedCallback):
         super(ProgressBarCallback, self).validation_start(dataloader)
         print("Running validation...")
         self.pbar = tqdm(total=len(dataloader), unit=" batches",
-                         desc="Validation {}".format(self.epoch + 1))
+                         desc="Validation {}".format(self.epoch))
 
-    def validation_step(self, batch, fwd_data, val_data):
+    def val_batch_end(self, batch, running_val_data):
         self.pbar.update(1)
 
     def validation_end(self, val_data):
@@ -378,13 +440,13 @@ class ProgressBarCallback(KeyedCallback):
         self.pbar.close()
         self.pbar = None
         s = " "*ProgressBarCallback.TABSTOPS + "Validation {} | ".format(
-            self.epoch + 1)
+            self.epoch)
         for k in self.val_keys:
             s += "{} = {:.2f} ".format(k, val_data[k])
         print(s)
 
-    def batch_end(self, batch_data, fwd, bwd_data):
-        super(ProgressBarCallback, self).batch_end(batch_data, fwd, bwd_data)
+    def batch_end(self, batch_data, train_step_data):
+        super(ProgressBarCallback, self).batch_end(batch_data, train_step_data)
         d = {}
         for k in self.keys:
             d[k] = self.ema[k]
@@ -399,8 +461,6 @@ class CheckpointingCallback(Callback):
 
     Args:
       checkpointer (Checkpointer): actual checkpointer responsible for the I/O.
-      start_epoch (int or None): index of the starting epoch (e.g. when
-          resuming from a previous checkpoint).
       interval (int, optional): minimum time in seconds between periodic
           checkpoints (within an epoch). There is not periodic checkpoint if
           this value is None.
@@ -413,7 +473,7 @@ class CheckpointingCallback(Callback):
     PERIODIC_PREFIX = "periodic_"
     EPOCH_PREFIX = "epoch_"
 
-    def __init__(self, checkpointer, start_epoch=None, interval=600,
+    def __init__(self, checkpointer, interval=600,
                  max_files=5, max_epochs=10):
         super(CheckpointingCallback, self).__init__()
         self.checkpointer = checkpointer
@@ -423,30 +483,22 @@ class CheckpointingCallback(Callback):
 
         self.last_checkpoint_time = time.time()
 
-        if start_epoch is None:
-            self.start_epoch = 0
-        else:
-            self.start_epoch = start_epoch
-
-    def get_extras(self):
-        return {"epoch": self.epoch + self.start_epoch}
-
     def epoch_end(self):
         """Save a checkpoint at the end of each epoch."""
         super(CheckpointingCallback, self).epoch_end()
         path = "{}{}".format(CheckpointingCallback.EPOCH_PREFIX, self.epoch)
-        self.checkpointer.save(path, extras=self.get_extras())
+        self.checkpointer.save(path, extras={"epoch": self.epoch + 1})
         self.__purge_old_files()
 
     def training_end(self):
         super(CheckpointingCallback, self).training_end()
-        self.checkpointer.save("training_end", extras=self.get_extras())
+        self.checkpointer.save("training_end", extras={"epoch": self.epoch + 1})
 
-    def batch_end(self, batch_data, fwd_result, bwd_result):
+    def batch_end(self, batch_data, train_step_data):
         """Save a periodic checkpoint if requested."""
 
         super(CheckpointingCallback, self).batch_end(
-            batch_data, fwd_result, bwd_result)
+            batch_data, train_step_data)
 
         if self.interval is None:  # We skip periodic checkpoints
             return
@@ -463,7 +515,7 @@ class CheckpointingCallback(Callback):
 
         filename = "{}{}".format(CheckpointingCallback.PERIODIC_PREFIX,
                                    time.strftime("%Y-%m-%d_%H-%M-%S", time.localtime()))
-        self.checkpointer.save(filename, extras=self.get_extras())
+        self.checkpointer.save(filename, extras={"epoch": self.epoch})
         self.__purge_old_files()
 
     def __purge_old_files(self):
@@ -518,107 +570,169 @@ class ImageDisplayCallback(Callback, abc.ABC):
         else:
             self.win = win
 
-    @abc.abstractmethod
-    def visualized_image(self, batch, fwd_result):
-        pass
+        self.first_validation_step = None
 
-    def caption(self, batch, fwd_result):
+    @abc.abstractmethod
+    def visualized_image(self, batch, step_data, is_val=False):
+        """Produces the image to visualize.
+
+        Args:
+            batch(Tensor, dict or tuple): batch of data.
+            batch(dict): batch of data produced by a ModelInterface's
+               `training_step` method (when is_val=True) or `validation_step`
+               when is_val=False.
+            is_val(bool): if True, the current visualization is for a
+                validation step.
+
+        Returns:
+            th.Tensor with shape [bs, 1 or 3, h, w] or None: image to display.
+
+        """
+        return None
+
+    def caption(self, batch, step_data, is_val=False):
         return ""
 
-    def batch_end(self, batch, fwd_result, bwd_result):
+    def batch_end(self, batch, step_data):
         if self._step % self.freq != 0:
             self._step += 1
             return
 
         self._step = 0
 
-        caption = self.caption(batch, fwd_result)
+        caption = self.caption(batch, step_data, is_val=False)
         opts = {"caption": "Epoch {}, batch {}: {}".format(
             self.epoch, self.batch, caption)}
 
-        viz = self.visualized_image(batch, fwd_result)
-        self._api.images(viz, win=self.win, opts=opts)
+        viz = self.visualized_image(batch, step_data, is_val=False)
+        if viz is not None:
+            self._api.images(viz, win=self.win, opts=opts)
         self._step += 1
 
     def validation_start(self, dataloader):
         super(ImageDisplayCallback, self).validation_start(dataloader)
-        self.first_step = True
+        self.first_validation_step = True
 
-    def validation_step(self, batch, fwd_data, val_data):
-        super(ImageDisplayCallback, self).validation_step(batch, fwd_data, val_data)
-        if not self.first_step:
+    def val_batch_end(self, batch, running_val_data):
+        super(ImageDisplayCallback, self).val_batch_end(batch, running_val_data)
+
+        # Only display the first validation image
+        if not self.first_validation_step:
             return
 
-        caption = self.caption(batch, fwd_data)
-        opts = {"caption": "Validation {}, batch {}: {}".format(
+        caption = self.caption(batch, running_val_data, is_val=True)
+        opts = {"caption": "Validation {}: {}".format(
             self.epoch, self.batch, caption)}
 
-        viz = self.visualized_image(batch, fwd_data)
-        self._api.images(viz, win=self.win+"_val", opts=opts)
-        self.first_step = False
+        viz = self.visualized_image(batch, running_val_data, is_val=True)
+        if viz is not None:
+            self._api.images(viz, win=self.win+"_val", opts=opts)
+        self.first_validation_step = False
 
 
-class ExperimentLoggerCallback(Callback):
-    """A callback that logs experiment parameters in a log."""
+class SQLLoggingCallback(KeyedCallback):
+    """A callback that logs scalar quantities to a .sqlite database.
+    """
+    RESERVED_KEYS = ["timestamp", "event", "step"]
 
-    def __init__(self, fname, meta=None):
-        super(ExperimentLoggerCallback, self).__init__()
-        LOG.error("ExperimentLoggerCallback is not implemented yet")
-        raise NotImplementedError("ExperimentLoggerCallback is not implemented yet")
+    def __init__(self, root, name=None, keys=None, val_keys=None,
+                 frequency=100, smoothing=0):
+        # No smoothing by default
+        super(SQLLoggingCallback, self).__init__(
+            keys=keys, val_keys=val_keys, smoothing=smoothing)
 
-    def training_start(self, dataloader):
-        super(ExperimentLoggerCallback, self).training_start(dataloader)
-        print("logging experiment with", self.datasize)
+        if name is None:
+            name = "logs"
+        name += ".sqlite"
 
-    def training_end(self):
-        super(ExperimentLoggerCallback, self).training_end()
-        print("end logging experiment", self.epoch, self.batch)
+        self.frequency = frequency
+        self.db = SQLiteDatabase(os.path.join(root, name))
+
+        for k in keys:
+            if k in SQLLoggingCallback.RESERVED_KEYS:
+                msg = "Key '%s' is reserved, please use a different name" % k
+                LOG.error(msg)
+                raise ValueError(msg)
+
+        # Counts the total number of optimization steps (batches)
+        self.step = 0
+        
+        previous = self.db.read_table("events")
+        if previous is not None:
+            # print("previous table", previous)
+            training_end = previous[previous["event"]=="training_end"]
+            last = training_end.tail(1)
+            if not last.empty:
+                self.step = last["step"].item()
+
+        # TODO: update database with new keys if they are missing
+
+        # TODO: index can be 'step'
 
     def _get_commit(self):
-       return subprocess.check_output(["git", "rev-parse", "HEAD"]) 
+       commit =  subprocess.check_output(["git", "rev-parse", "HEAD"])
+       return commit.strip()
 
+    def _save_to_db(self, data, table="logs"):
+        current_time = datetime.datetime.now()
+        data["timestamp"] = current_time
+        data["step"] = self.step
+        self.db.append_row(data, table)
 
-class CSVLoggingCallback(KeyedCallback):
-    """A callback that logs scalar quantities to a .csv file.
+    def batch_end(self, batch_data, train_step_data):
+        """Logs training advancement batch"""
+        super(SQLLoggingCallback, self).batch_end(batch_data, train_step_data)
 
-    Format is:
-        epoch, step, event, key, value
-    """
-    def __init__(self, fname, keys=None, val_keys=None, smoothing=0):
-        super(CSVLoggingCallback, self).__init__(keys=keys, val_keys=val_keys, smoothing=smoothing)
+        # Allow subsampling of the data
+        if self.step % self.frequency == 0:
+            data = {k: train_step_data[k] for k in self.keys}
+            self._save_to_db(data)
+            print("save", self.step)
 
-        LOG.error("CSVLoggingCallback is not implemented yet")
-        raise NotImplementedError("CSVLoggingCallback is not implemented yet")
-
-        self.fname = fname
-        self.fid = open(self.fname, 'w')
-
-        self.fid.write("epoch, step, event, key, value\n")
-        self.fid.write(",,logger_created,,\n")
-
-
-        # open file, check last event
-
-    def __del__(self):
-        LOG.info("deleting csv logger")
-        self.fid.write(",,logger_deleted,,\n")
-        self.fid.close()
-
-    def batch_end(self, batch_data, fwd, bwd_data):
-        """Logs training advancement Batch"""
-        super(CSVLoggingCallback, self).batch_end(batch_data, fwd, bwd_data)
-
-        for k in self.keys:
-            v = bwd_data[k]
-            self.fid.write("%d,%d,batch_end,%s,%f\n" % (self.epoch, self.batch, k, v))
+        self.step += 1
 
     def training_start(self, dataloader):
-        super(CSVLoggingCallback, self).training_start(dataloader)
-        self.fid.write(",,training_start,,\n")
+        super(SQLLoggingCallback, self).training_start(dataloader)
+        data = {
+            "event": "training_start",
+            "git_commit": self._get_commit(),
+        }
+        self._save_to_db(data, table="events")
 
     def training_end(self):
-        super(CSVLoggingCallback, self).training_end()
-        self.fid.write(",,training_end,,\n")
+        super(SQLLoggingCallback, self).training_end()
+        data = {
+            "event": "training_end",
+            "git_commit": None,
+        }
+        self._save_to_db(data, table="events")
+
+    def validation_end(self, val_data):
+        super(SQLLoggingCallback, self).validation_end(val_data)
+        data = {k: val_data[k] for k in self.val_keys}
+        self._save_to_db(data, table="val_logs")
+
+        data = {
+            "event": "validation",
+            "git_commit": None,
+        }
+        self._save_to_db(data, table="events")
+
+    def epoch_start(self, epoch):
+        super(SQLLoggingCallback, self).epoch_start(epoch)
+        data = {
+            "event": "epoch_start",
+            "git_commit": None,
+        }
+        self._save_to_db(data, table="events")
+
+    def epoch_end(self):
+        super(SQLLoggingCallback, self).epoch_end()
+        data = {
+            "event": "epoch_end",
+            "git_commit": None,
+        }
+        self._save_to_db(data, table="events")
 
 
 def _random_string(size=16):
@@ -646,8 +760,8 @@ class TensorBoardLoggingCallback(Callback):
         self.frequency = frequency
         self.summary_type = summary_type
 
-    def batch_end(self, batch_data, fwd, bwd):
-        super(TensorBoardLoggingCallback, self).batch_end(batch_data, fwd, bwd)
+    def batch_end(self, batch_data, train_step_data):
+        super(TensorBoardLoggingCallback, self).batch_end(batch_data, train_step_data)
 
         if self._step % self.frequency != 0:
             self._step += 1
@@ -658,9 +772,9 @@ class TensorBoardLoggingCallback(Callback):
 
         for k in self.keys:
             if self.summary_type == 'scalar':
-                self._writer.add_scalar(k, bwd[k], global_step=t)
+                self._writer.add_scalar(k, train_step_data[k], global_step=t)
             elif self.summary_type == 'histogram':
-                self._writer.add_histogram(k, bwd[k], global_step=t)
+                self._writer.add_histogram(k, train_step_data[k], global_step=t)
         self._step += 1
 
     def validation_end(self, val_data):
@@ -691,21 +805,21 @@ class TensorBoardImageDisplayCallback(Callback, abc.ABC):
         self._step = 0
 
     @abc.abstractmethod
-    def visualized_image(self, batch, fwd_result):
+    def visualized_image(self, batch, train_step_data):
         pass
 
     @abc.abstractmethod
     def tag(self):
         pass
 
-    def batch_end(self, batch, fwd_result, bwd_result):
+    def batch_end(self, batch, train_step_data):
         if self._step % self.freq != 0:
             self._step += 1
             return
 
         self._step = 0
 
-        viz = self.visualized_image(batch, fwd_result)
+        viz = self.visualized_image(batch, train_step_data)
         t = self.batch + self.datasize * self.epoch
         self._writer.add_image(self.tag(), make_grid(viz), t)
         self._step += 1
@@ -714,12 +828,28 @@ class TensorBoardImageDisplayCallback(Callback, abc.ABC):
         super(TensorBoardImageDisplayCallback, self).validation_start(dataloader)
         self.first_step = True
 
-    def validation_step(self, batch, fwd_data, val_data):
-        super(TensorBoardImageDisplayCallback, self).validation_step(batch, fwd_data, val_data)
+    def val_batch_end(self, batch, running_val_data):
+        super(TensorBoardImageDisplayCallback, self).val_batch_end(batch, running_val_data)
         if not self.first_step:
             return
 
-        viz = self.visualized_image(batch, fwd_data)
+        viz = self.visualized_image(batch, running_val_data)
         t = self.datasize * (self.epoch+1)
         self._val_writer.add_image(self.tag(), make_grid(viz), t)
         self.first_step = False
+
+
+class LRSchedulerCallback(Callback):
+    """
+    Args:
+        schedulers: th.optim.Scheduler or list.
+    """
+    def __init__(self, schedulers):
+        # Make it a list
+        if not isinstance(schedulers, list):
+            schedulers = [schedulers]
+        self.schedulers = schedulers
+
+    def epoch_end(self):
+        for s in self.schedulers:
+            s.step()
