@@ -64,35 +64,20 @@ class ModelInterface(metaclass=ABCMeta):
         pass
 
     @abstractmethod
-    def forward(self, batch):
-        """Runs the model on a batch of data.
+    def training_step(self, batch):
+        """Training step given a batch of data.
+
+        This should implement a forward pass of the model, compute gradients,
+        take an optimizer step and return useful metrics and tensors for
+        visualization and training callbacks. 
 
         Args:
           batch (dict): batch of data provided by a data pipeline.
 
         Returns:
-          forward_data (dict): a dictionary of outputs
+          train_step_data (dict): a dictionary of outputs.
         """
-
-        forward_data = {}
-
-        return forward_data
-
-    @abstractmethod
-    def backward(self, batch, forward_data):
-        """Computes gradients, take an optimizer step and update the model.
-
-        Args:
-          batch (dict): batch of data provided by a data pipeline.
-          forward_data (dict): outputs from the forward pass
-
-        Returns:
-          backward_data (dict): a dictionary of outputs
-        """
-
-        backward_data = {}
-
-        return backward_data
+        return {}
 
     def init_validation(self):
         """Initializes the quantities to be reported during validation.
@@ -102,40 +87,24 @@ class ModelInterface(metaclass=ABCMeta):
         Returns:
           data (dict): initialized values
         """
-
+        LOG.warning("Running a ModelInterface validation initialization that was not overriden: this is a no-op.")
         data = {}
         return data
 
-    def update_validation(self, batch, fwd, running_data):
-        """Updates the running val data using the current batch's forward output.
+    def validation_step(self, batch, running_val_data):
+        """Updates the running validataion with the current batch's results.
 
         The default implementation is a no-op
 
         Args:
           batch (dict): batch of data provided by a data pipeline.
-          fwd (dict): data from one forward step in validation mode
-          running_data (dict): current aggregates of the validation loop.
+          running_val_data (dict): current aggregates of the validation loop.
 
         Returns:
-          updated_data (dict): initialized values
+          updated_data (dict): new updated value for the running_val_data.
         """
-
-        updated_data = {}
-        return updated_data
-
-    def finalize_validation(self, running_data):
-        """Computes the final validation aggregates from the running data.
-
-        The default implementation is a no-op
-
-        Args:
-          running_data (dict): current aggregates of the validation loop.
-
-        Returns:
-          validation_data (dict): initialized values
-        """
-
-        return running_data
+        LOG.warning("Running a ModelInterface validation step that was not overriden: this is a no-op.")
+        return {}
 
     def __repr__(self):
         return self.__class__.__name__
@@ -164,6 +133,7 @@ class Trainer(object):
         self._keep_running = True
 
     def interrupt_handler(self, signo, frame):
+        """Stop the training process upon receiving a SIGINT (Ctrl+C)."""
         LOG.debug("interrupting run")
         self._keep_running = False
 
@@ -173,46 +143,60 @@ class Trainer(object):
         self.__training_end()
 
     def add_callback(self, callback):
-        """Adds a callback to the list of training hooks."""
+        """Adds a callback to the list of training hooks.
+
+        Args:
+            callback(ttools.Callback): callback to add.
+        """
         LOG.debug("Adding callback {}".format(callback))
+        # pass an interface reference to the callback
+        callback.model_interface = self.interface
         self.callbacks.append(callback)
 
-    def train(self, dataloader, num_epochs=None, val_dataloader=None):
+    def train(self, dataloader, starting_epoch=None, num_epochs=None,
+              val_dataloader=None):
         """Main training loop. This starts the training procedure.
 
         Args:
           dataloader (DataLoader): loader that yields training batches.
+          starting_epoch (int, optional): index of the epoch we are starting from.
           num_epochs (int, optional): max number of epochs to run.
           val_dataloader (DataLoader, optional): loader that yields validation
             batches
         """
         self.__training_start(dataloader)
-        epoch = 0
-        while num_epochs is None or epoch < num_epochs:
+        if starting_epoch is None:
+            starting_epoch = 0
+
+        LOG.info("Starting taining from epoch %d", starting_epoch)
+
+        epoch = starting_epoch
+
+        while num_epochs is None or epoch < starting_epoch + num_epochs:
             self.__epoch_start(epoch)
             for batch_idx, batch in enumerate(dataloader):
                 if not self._keep_running:
                     self._stop()
                     return
                 self.__batch_start(batch_idx, batch)
-                fwd_result = self.__forward_step(batch)
-                bwd_result = self.__backward_step(batch, fwd_result)
-                self.__batch_end(batch, fwd_result, bwd_result)
+                train_step_data = self.__training_step(batch)
+                self.__batch_end(batch, train_step_data)
             self.__epoch_end()
 
+            # TODO: allow validation at intermediate steps during one epoch
+
             # Validate
             if val_dataloader:
                 with th.no_grad():
-                    val_data = self.__validation_start(
-                        val_dataloader)  # data interface adapter
+                    running_val_data = self.__validation_start(val_dataloader)
                     for batch_idx, batch in enumerate(val_dataloader):
                         if not self._keep_running:
                             self._stop()
                             return
-                        fwd_result = self.__forward_step(batch)
-                        val_data = self.__validation_update(
-                            batch, fwd_result, val_data)
-                    self.__validation_end(val_data)
+                        self.__val_batch_start(batch_idx, batch)
+                        running_val_data = self.__validation_step(batch, running_val_data)
+                        self.__val_batch_end(batch, running_val_data)
+                    self.__validation_end(running_val_data)
 
             epoch += 1
 
@@ -226,12 +210,6 @@ class Trainer(object):
         return "Trainer({}, {} callbacks)".format(
             self.interface, len(self.callbacks))
 
-    def __forward_step(self, batch):
-        return self.interface.forward(batch)
-
-    def __backward_step(self, batch, forward_data):
-        return self.interface.backward(batch, forward_data)
-
     def __training_start(self, dataloader):
         for cb in self.callbacks:
             cb.training_start(dataloader)
@@ -248,28 +226,36 @@ class Trainer(object):
         for cb in self.callbacks:
             cb.epoch_end()
 
+    def __batch_start(self, batch_idx, batch):
+        for cb in self.callbacks:
+            cb.batch_start(batch_idx, batch)
+
+    def __batch_end(self, batch, train_step_data):
+        for cb in self.callbacks:
+            cb.batch_end(batch, train_step_data)
+
+    def __val_batch_start(self, batch_idx, batch):
+        for cb in self.callbacks:
+            cb.val_batch_start(batch_idx, batch)
+
+    def __val_batch_end(self, batch, running_val_data):
+        for cb in self.callbacks:
+            cb.val_batch_end(batch, running_val_data)
+
     def __validation_start(self, dataloader):
         for cb in self.callbacks:
             cb.validation_start(dataloader)
         return self.interface.init_validation()
 
-    def __validation_update(self, batch, fwd_data, val_data):
+    def __validation_end(self, running_val_data):
         for cb in self.callbacks:
-            cb.validation_step(batch, fwd_data, val_data)
-        return self.interface.update_validation(batch, fwd_data, val_data)
+            cb.validation_end(running_val_data)
 
-    def __validation_end(self, val_data):
-        val_data = self.interface.finalize_validation(val_data)
-        for cb in self.callbacks:
-            cb.validation_end(val_data)
+    def __training_step(self, batch):
+        return self.interface.training_step(batch)
 
-    def __batch_start(self, batch_idx, batch):
-        for cb in self.callbacks:
-            cb.batch_start(batch_idx, batch)
-
-    def __batch_end(self, batch, fwd_result, bwd_result):
-        for cb in self.callbacks:
-            cb.batch_end(batch, fwd_result, bwd_result)
+    def __validation_step(self, batch, running_val_data):
+        return self.interface.validation_step(batch, running_val_data)
 
 
 class Checkpointer(object):
@@ -278,21 +264,40 @@ class Checkpointer(object):
     Args:
       root (string): path to the root directory where the files are stored.
       model (torch.nn.Module):
-      optimizers (list of torch.optimizer): optimizers whose parameters will
+      meta (dict): a dictionary of training or configuration parameters useful
+          to initialize the model upon loading the checkpoint again.
+      optimizers (single or list of torch.optimizer): optimizers whose parameters will
         be checkpointed together with the model.
-      meta (dict):
+      schedulers (single or list of
+      torch.optim.lr_scheduler): schedulers whose
+          parameters will be checkpointed together with
+          the model.
+      prefix (str): unique prefix name in case several models are stored in the
+        same folder.
     """
 
     EXTENSION = ".pth"
 
-    def __init__(self, root, model=None, meta=None, optimizers=None, prefix=None):
+    def __init__(self, root, model=None, meta=None, optimizers=None,
+                 schedulers=None, prefix=None):
         self.root = root
         self.model = model
         self.meta = meta
 
-        if optimizers is not None and not isinstance(optimizers, list):
-            raise ValueError("optimizers should be a list or None")
+        # TODO(mgharbi): verify the prefixes are unique.
+
+        if optimizers is None:
+            LOG.info("No optimizer state will be stored in the "
+                        "checkpointer")
+        else:
+            # if we have only one optimizer, make it a list
+            if not isinstance(optimizers, list):
+                optimizers = [optimizers]
         self.optimizers = optimizers
+        if schedulers is not None:
+            if not isinstance(schedulers, list):
+                schedulers = [schedulers]
+        self.schedulers = schedulers
 
         LOG.debug(self)
 
@@ -327,22 +332,36 @@ class Checkpointer(object):
             for opt in self.optimizers:
                 opt_dicts.append(opt.state_dict())
 
+        sched_dicts = []
+        if self.schedulers is not None:
+            for s in self.schedulers:
+                sched_dicts.append(s.state_dict())
+
         filename = self.__path(path, prefix=self.prefix)
         os.makedirs(self.root, exist_ok=True)
         th.save({'model': model_state,
                  'meta': self.meta,
                  'extras': extras,
                  'optimizers': opt_dicts,
+                 'schedulers': sched_dicts,
                  }, filename)
         LOG.debug("Checkpoint saved to \"{}\"".format(filename))
 
     def try_and_init_from(self, path):
+        """Try to initialize the models's weights from an external checkpoint.
+
+        Args:
+            path(str): full path to the checkpoints to load model parameters
+                from.
+        """
         LOG.info("Loading weights from foreign checkpoint {}".format(path))
         if not os.path.exists(path):
             raise ValueError("Checkpoint {} does not exist".format(path))
+
         chkpt = th.load(path, map_location=th.device("cpu"))
         if "model" not in chkpt.keys() or chkpt["model"] is None:
             raise ValueError("{} has no model saved".format(path))
+
         mdl = chkpt["model"]
         for n, p in self.model.named_parameters():
             if n in mdl:
@@ -354,6 +373,7 @@ class Checkpointer(object):
                 p.data.copy_(p2)
             else:
                 LOG.warning("Parameter {} ignored, not found in source checkpoint.".format(n))
+
         LOG.info("Weights loaded from foreign checkpoint {}".format(path))
 
     def load(self, path):
@@ -377,12 +397,26 @@ class Checkpointer(object):
         if "optimizers" in chkpt.keys():
             if self.optimizers is not None and chkpt["optimizers"] is not None:
                 try:
-                    for opt, state in zip(self.optimizers, chkpt["optimizers"]):
+                    for opt, state in zip(self.optimizers,
+                                          chkpt["optimizers"]):
                         LOG.debug("Loading optimizers state dict for %s", opt)
                         opt.load_state_dict(state)
                 except:
-                    # We do raise an error here, e.g. in case the user simply changes optimizer
-                    LOG.warning("Could not load optimizer state dicts, starting from scratch")
+                    # We do not raise an error here, e.g. in case the user simply
+                    # changes optimizer
+                    LOG.warning("Could not load optimizer state dicts, "
+                                "starting from scratch")
+
+        if "schedulers" in chkpt.keys():
+            if self.schedulers is not None and chkpt["schedulers"] is not None:
+                try:
+                    for s, state in zip(self.schedulers,
+                                          chkpt["schedulers"]):
+                        LOG.debug("Loading scheduler state dict for %s", s)
+                        s.load_state_dict(state)
+                except:
+                    LOG.warning("Could not load scheduler state dicts, "
+                                "starting from scratch")
 
         LOG.debug("Loaded checkpoint \"{}\"".format(filename))
         return tuple(chkpt[k] for k in ["extras", "meta"])
@@ -391,7 +425,8 @@ class Checkpointer(object):
         """Try to load the most recent checkpoint, skip failing files.
 
         Returns:
-          extras (dict): extra information passed by the user at save time.
+          extras (dict): extra user-defined information that was saved in the
+              checkpoint.
           meta (dict): metaparameters of the model passed at save time.
         """
         all_checkpoints = self.sorted_checkpoints()
@@ -409,7 +444,11 @@ class Checkpointer(object):
         return extras, meta
 
     def sorted_checkpoints(self):
-        """Get list of all checkpoints in root directory, sorted by creation date."""
+        """Get list of all checkpoints in root directory, sorted by creation date.
+
+        Returns:
+            chkpts (list of str): sorted checkpoints in the root folder.
+        """
         reg = re.compile(r"{}.*\{}".format(self.prefix, Checkpointer.EXTENSION))
         if not os.path.exists(self.root):
             all_checkpoints = []
@@ -426,7 +465,11 @@ class Checkpointer(object):
         return chkpts
 
     def delete(self, path):
-        """Delete checkpoint at path."""
+        """Delete checkpoint at path.
+
+        Args:
+            path(str): full path to the checkpoint to delete.
+        """
         if path in self.sorted_checkpoints():
             os.remove(os.path.join(self.root, path))
         else:
@@ -434,8 +477,18 @@ class Checkpointer(object):
 
     @staticmethod
     def load_meta(root, prefix=None):
-        """Fetch model metadata without touching the saved parameters."""
-        chkptr = Checkpointer(root, model=None, meta=None, prefix=prefix)
+        """Fetch model metadata without touching the saved parameters.
+
+        This loads the metadata from the most recent checkpoint in the root
+        directory.
+
+        Args:
+            root(str): path to the root directory containing the checkpoints
+            prefix(str): unique prefix for the checkpoint to be loaded (e.g. if
+                multiple models are saved in the same location)
+        """
+        chkptr = Checkpointer(root, model=None, meta=None, prefix=prefix, 
+                              optimizers=[])
         LOG.debug("checkpoints: %s", chkptr.sorted_checkpoints())
         _, meta = chkptr.load_latest()
         return meta
